@@ -12,7 +12,7 @@ Dancer::Plugin::Database - easy database connections for Dancer applications
 
 =cut
 
-our $VERSION = '1.30';
+our $VERSION = '1.40';
 
 my $settings = undef;
 
@@ -55,8 +55,14 @@ register database => sub {
         }
     }
 
+    # To be fork safe and thread safe, use a combination of the PID and TID (if
+    # running with use threads) to make sure no two processes/threads share
+    # handles.  Implementation based on DBIx::Connector by David E. Wheeler.
+    my $pid_tid = $$;
+    $pid_tid .= '_' . threads->tid if $INC{'threads.pm'};
+
     # OK, see if we have a matching handle
-    $handle = $handles{$handle_key} || {};
+    $handle = $handles{$pid_tid}{$handle_key} || {};
     
     if ($handle->{dbh}) {
         if ($conn_details->{connection_check_threshold} &&
@@ -81,7 +87,7 @@ register database => sub {
         # Get a new connection
         if ($handle->{dbh} = _get_connection($conn_details)) {
             $handle->{last_connection_check} = time;
-            $handles{$handle_key} = $handle;
+            $handles{$pid_tid}{$handle_key} = $handle;
             return $handle->{dbh};
         } else {
             return;
@@ -97,10 +103,13 @@ sub _get_connection {
 
     # Assemble the DSN:
     my $dsn;
+    my $driver;
     if ($settings->{dsn}) {
         $dsn = $settings->{dsn};
+        ($driver) = $dsn =~ m{dbi:([^:]+)};
     } else {
         $dsn = "dbi:" . $settings->{driver};
+        $driver = $settings->{driver};
         my @extra_args;
 
         # DBD::SQLite wants 'dbname', not 'database', so special-case this
@@ -109,7 +118,7 @@ sub _get_connection {
         # things easier for our users by handling this for them):
         # (DBD::SQLite will support 'database', too, as of 1.32 when it's
         # released)
-        if ($settings->{driver} eq 'SQLite' 
+        if ($driver eq 'SQLite' 
             && $settings->{database} && !$settings->{dbname}) {
             $settings->{dbname} = delete $settings->{database};
         }
@@ -135,7 +144,7 @@ sub _get_connection {
             mysql  => 'mysql_enable_utf8',
             Pg     => 'pg_enable_utf8',
         );
-        if (my $param = $param_for_driver{ $settings->{driver} }) {
+        if (my $param = $param_for_driver{$driver}) {
             Dancer::Logger::debug(
                 "Adding $param to DBI connection params to enable UTF-8 support"
             );
@@ -289,6 +298,11 @@ no-op query against the database if not.  If the connection has gone away, a new
 connection will be obtained and returned.  This avoids any problems for
 a long-running script where the connection to the database might go away.
 
+Care is taken that handles are not shared across processes/threads, so this
+should be thread-safe with no issues with transactions etc.  (Thanks to Matt S
+Trout for pointing out the previous lack of thread safety.  Inspiration was
+drawn from DBIx::Connector.)
+
 =head1 CONFIGURATION
 
 Connection details will be taken from your Dancer application config file, and
@@ -461,6 +475,11 @@ Martin J Evans
 
 Carlos Sosa
 
+Matt S Trout
+
+Matthew Vickers
+
+
 =head1 BUGS
 
 Please report any bugs or feature requests to C<bug-dancer-plugin-database at rt.cpan.org>, or through
